/*
This file is part of "PMR466 intercom".

Copyright 2021 by M.T. Konstapel
https://www.meezenest.nl/mees

Version 1 - 2021-02-19
First working release

Version 1.1 - 2021-02-19
Initialization after reset: both LEDs and the 12V power supply are off (was on)

"PMR intercom" is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

"PMR466 intercom" is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with "PMR446 intercom" .  If not, see <http://www.gnu.org/licenses/>.
*/

#include "fuses.h"
#pragma chip SX28
#pragma FUSE = TURBO & WDT_OFF & FOSC_HS1
#pragma FUSEX = STACKX & CF_1
#define Clock_4MHz

#include "math24.h"

#define TRUE 1
#define FALSE 0

#define LOW   0
#define HIGH  1

bit ON_SIG      @ RA.0;              //Signals external device when intercom is active
bit LED         @ RA.1;              //LED on outside intercom module
bit STATUS_LED  @ RA.2;              //Status LED on control unit
bit SWITCH      @ RB.0;              //BELL switch on outside intercom module
bit PS_ENA      @ RB.1;              //Power supply enable signal (input)
bit PS_ON       @ RB.5;              //Switches 12V power supply on or off
bit BELL        @ RC.5;              //Activates solenoid of copper bell
bit PTT         @ RC.2;              //Push to talk

bit BEEP0       @ RC.3;              //Dial tone output 0
bit BEEP1       @ RC.4;              //Dial tone output 1

#pragma codepage 0
#pragma rambank 0

#define MAX_NR_OF_BEEPS 4
#define DEBOUNCE_DELAY  150          // Debounce time of approximately 56 msec

uns8 SolenoidCounter;                // Holds track of the number of times the solenoid is being activated

// Interrupt from RTCC
// This ISR generates an old school dial tone like sound when RTCC interrupt is activated
// It also blinks both LEDs
uns16 BeepCounter;                   // Counter for 320Hz, 1.25Hz and 0.3125Hz generators
uns8 BeepCounter2;                   // Counter for 440 Hz generator
uns16 ISRCounter;
uns16 UptimeCounter;                 // Counter for uptime, increments every 24.6 seconds
bit UptimeClockTick;
bit StartDialTone;                   // Set to TRUE generates dial tone
interrupt int_server (void)
// located at address 0
{
    // W, STATUS and FSR are automatically saved

    RTCC = 243;        //With prescaler = 128 and chrystal = 4.433619 MHz: interrupt rate = 2664 Hz.

    // Set UptimeCounterTick when counter rolls over
    if (++ISRCounter == 0xFFFF)
       UptimeClockTick = TRUE;

    // Generate dial tone and flash som LEDs
    // Dial tone is (approximately) 0.4 second beep, 0.4 second silence, 0.4 second beep, 1.6 second silence, than repeat
    // beep consists of two signals, one 330 Hz and one 440Hz, resulting in beat of 110Hz
    if (StartDialTone) {
        BeepCounter++;
        // Generate two short 330Hz pulses
        if ( (BeepCounter & 0x04) && !(BeepCounter & 0x400) && !(BeepCounter & 0x1000))
           BEEP0 = HIGH;
        else
           BEEP0 = LOW;

           // Generate two short 440Hz pulses
           if ((++BeepCounter2 == 3) && !(BeepCounter & 0x400) && !(BeepCounter & 0x1000) ) {
           BEEP1 ^= 1;
           BeepCounter2 = 0;

        }
        // Pulse both LED on intercom and STATUS_LED on controller
        // LED = LOW ==> LED is on
        if ( !(BeepCounter & 0x1000) )  {
           LED = HIGH;
           STATUS_LED = LOW;
        } else {
           LED = LOW;
           STATUS_LED = HIGH;
        }
    }

}

//Initialize sleep mode
//Wake up from sleep is the same as reset
void PowerDown(void)
{
    // Set all pins to default before entering sleep mode
    ON_SIG     = HIGH;    // ON_SIGNAL off
    LED        = LOW;     // LED on intercom module off
    STATUS_LED = HIGH;    // STATUS LED off
    PS_ON      = LOW;     // 12 Volt power supply off
    BELL       = LOW;     // Solenoid BELL not activated
    PTT        = LOW;     // No PTT signal
    BEEP0      = LOW;     // 0 Volt on BEEP signal
    BEEP1      = LOW;     // 0 Volt on BEEP signal

    WKED_B = 0x01;        //Interrupt on falling edge of pin 0 (SWITCH)
    WKPND_B = 0x0;        // clear interrupt pending bits
    WKEN_B = 0xFE;        //clear bit 0, this enables wakeup on interrupt on pin 0 (SWITCH)

    sleep();              //Go to sleep
}

void InitSystem (void)
{
    // Initialize output pins
    ON_SIG     = HIGH;    // ON_SIGNAL off
    LED        = LOW;     // LED on intercom module off
    STATUS_LED = HIGH;    // STATUS LED off
    PS_ON      = LOW;     // 12 Volt power supply off
    BELL       = LOW;     // Solenoid BELL not activated
    PTT        = LOW;     // No PTT signal
    BEEP0      = LOW;     // 0 Volt on BEEP signal
    BEEP1      = LOW;     // 0 Volt on BEEP signal

    DDRA  = 0x08;         // RA0...RA2 output
    DDRB  = 0xDF;         // RB0 and RB1 input, RB4 output, others unused (set is input)
    PLP_B = 0xFC;         // Pullup on RB0 and RB1 input
    DDRC  = 0xC3;         // RC2...RC5 output, others unused (set as input)

    BeepCounter=0;        // Counter for 320Hz, 1.25Hz and 0.3125Hz generators
    BeepCounter2=0;       // Counter for 440 Hz generator
    ISRCounter=0;         // Counter for uptime, increments on ISR
    UptimeClockTick = FALSE; // ISR sets this flag every 24.6 seconds
    UptimeCounter=0;      // 24.6 second counter

    SolenoidCounter = 0;  // Reset SolenoidCounter

    StartDialTone = FALSE; // Generate dial tone

    // Initialize RTCC and interrupt
    WKPND_B = 0x0;        // clear interrupt pending bits
    OPTION  = 0x86;       // RTCC interrupt enabled, internal clock, with prescaler, divider ration 1:128
                          // With 4.433619MHz chrystal this gives a RTCC clock rate of 34.638 kHz
}

void ActivateSolenoid(void)
{

     if (ISRCounter > 0x200) {
        ISRCounter = 0;

        switch (SolenoidCounter){
             case 0:
               BELL = HIGH;
               SolenoidCounter++;
               break;
             case 1:
               BELL = LOW;
               SolenoidCounter++;
               break;
             case 2:
               BELL = HIGH;
               SolenoidCounter++;
               break;
             case 3:
               BELL = LOW;
               SolenoidCounter++;
             default:
               break;
             }

        }
}

//#pragma codepage 1
#pragma rambank 1
uns24 DebounceTimer;             //24 bit should be enough as UptimeCounter is never more than 4
uns24 LastDebounceTimer;
bit CurrentButtonState;
bit SampleButton(void)
{
     bit ButtonPressed;

     CurrentButtonState = !SWITCH;      // Sample switch

     if (CurrentButtonState) {          // Switch pressed?

          // Debounce button
          LastDebounceTimer = DebounceTimer;
          DebounceTimer = ( (uns24)UptimeCounter + (uns24)ISRCounter ) - LastDebounceTimer;

          if (DebounceTimer > DEBOUNCE_DELAY)
             ButtonPressed = TRUE;

     } else {

           DebounceTimer = (uns24)UptimeCounter + (uns24)ISRCounter;
           ButtonPressed = FALSE;

     }

     return (ButtonPressed);
}

//#pragma codepage 0
#pragma rambank 0

void main( void)
{
    bit JustAwake = TRUE;      // This bit is TRUE after RESET, then it is set to FALSE;
    uns8 LED_is_ON = FALSE;    // Used for counting number of beeps
    uns8 CountNrOfBeeps = 0;   // Counts number of beeps generated by the ISR

    InitSystem();

    while (1)  {  /* infinite loop */

        if ( SampleButton() ) {

           // Is it the first time since a reset?
           if (JustAwake) {

              ON_SIG     = LOW;     // ON_SIGNAL on
              LED        = HIGH;    // LED on intercom module on
              STATUS_LED = LOW;     // STATUS LED on
              PS_ON      = HIGH;    // 12 Volt power supply on

              StartDialTone = TRUE;  // Start ringing the bell
              JustAwake = FALSE;

           // Button pressed while intercom is on: SWITCH is now PTT
           } else if (!StartDialTone) {
             PTT = HIGH;               // Start transmitting
             UptimeCounter = 0;        // Reset UptimeCounter, so we don't fall asleep during a conversation
           }
       } else if (!StartDialTone) {   // SWITCH is not pressed and we are not ringing the bell any more
         PTT = LOW;                   // Stop transmitting
       }

       // The ISR is generating a dial tone. We should do some important stuff now!
       if (StartDialTone) {

          // Put the transceiver in TX mode
          PTT = HIGH;

          // Ring the old school bell
          ActivateSolenoid();

          // Make shure there are only MAX_NR_OF_BEEPS generated
          // We can count the number of beebs generated by the ISR by counting the pulses on the LED
          if (LED)
             LED_is_ON = TRUE;

          else {

            if (LED_is_ON) {
               LED_is_ON = FALSE;
               CountNrOfBeeps++;
            }

          }

          if (CountNrOfBeeps >= MAX_NR_OF_BEEPS) {
             StartDialTone = FALSE;     // Stop generating dial tone
             CountNrOfBeeps = 0;
             PTT = LOW;                 // Stop transmitting
             STATUS_LED = LOW;          // LEDs continuous on
             LED = HIGH;
          }
       }

       // Update UptimeCounter
       if (UptimeClockTick) {
          UptimeClockTick = FALSE;
          UptimeCounter++;
       }

       // After 50-75 seconds go to low power mode (sleep)
       if (UptimeCounter > 3)
          PowerDown();

    }

}
